from . import constant, url
from .attribute import FetchedResponses, Schedule, Student, Token, User
from .type import (Any, ClientType, Dict, Enum, EnumMeta, Optional, Response,
                   Union)


def response_to_json(response: Union[Response, dict]) -> dict:
    '''Convert response to json. Do nothing if response is already dictionary

    Parameters
    ----------
    response : Union[Response, dict]
        The response to convert.

    Returns
    -------
    dict
        Dictionary representation of the response.

    Raises
    ------
    TypeError
        Invalid response type.
    '''

    if isinstance(response, dict):
        return response

    if isinstance(response, Response):
        return response.json()

    raise TypeError("response must be Response or dict")


def extract(response: Union[Response, dict], attr: Enum) -> Any:
    '''Extract any value from login response or schedule response.
    Use enums from pymyku.attribute as key to get value.

    Parameters
    ----------
    response : Union[Response, dict]
        The response to extract from. (Can be login response or schedule response)
    attr : Enum
        The attribute to extract.

    Returns
    -------
    Any
        Value of the attribute.

    Raises
    ------
    TypeError
        Invalid attr type.
    '''

    if isinstance(attr, EnumMeta):
        raise TypeError(
            "attr must be Enum not EnumMeta. Use '.' operator after the EnumMeta object.")

    if not isinstance(attr, Enum):
        raise TypeError("attr must be Enum.")

    response = response_to_json(response)

    if isinstance(attr, User):
        return response.get("user", {})\
            .get(attr.value, None)

    if isinstance(attr, Student):
        return response.get("user", {})\
            .get("student", {})\
            .get(attr.value, None)

    if isinstance(attr, Token):
        return response.get(attr.value, None)

    if isinstance(attr, Schedule):
        result = response.get("results", [])
        if not result:
            return None
        return result[0].get(attr.value, None)


def get_raise(dictionary: dict, key: str, dict_name: Optional[str] = None) -> Any:
    '''Raise ValueError if key is not found in dictionary.

    Parameters
    ----------
    dictionary : dict
        The dictionary to search in.
    key : str
        The key to search for.
    dict_name : Optional[str]
        Name of the dictionary.

    Returns
    -------
    Any
        Value of the key.

    Raises
    ------
    ValueError
        If key is not found in dictionary.
    '''

    result = dictionary.get(key)
    if result is None:
        raise ValueError(
            f'{key} is not found in {dict_name if dict_name else "dictionary"}')
    return result


def extract_user_data(login_response: Union[Response, dict]) -> dict:
    '''Extract user data from login response.

    Parameters
    ----------
    login_response : Union[Response, dict]
        The response of the login request.

    Returns
    -------
    dict
        User data. Represented by :class:`pymyku.attribute.User`
        
    Raises
    ------
    ValueError
        If user data is not found.
    '''

    login_response = response_to_json(login_response)

    result = get_raise(login_response, 'user', 'login response')

    return result


def extract_student_data(login_response: Union[Response, dict]) -> dict:
    '''Extract student data from login response.

    Parameters
    ----------
    login_response : Union[Response, dict]
        The response of the login request.

    Returns
    -------
    dict
        Student data. Represented by :class:`pymyku.attribute.Student`
        
    Raises
    ------
    ValueError
        If student data is not found.
    '''

    user_data = extract_user_data(login_response)

    result = get_raise(user_data, 'student', 'login response')

    return result


def extract_access_token(login_response: Union[Response, dict]) -> str:
    '''Extract student code from login response

    Parameters
    ----------
    login_response : Union[Response, dict]
        This is the response from the login request.

    Returns
    -------
    str
        Access token. Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
        
    Raises
    ------
    ValueError
        If access token is not found.
    '''

    login_response = response_to_json(login_response)

    result = get_raise(login_response, 'accesstoken', 'login response')

    return result


def extract_std_code(login_response: Union[Response, dict]) -> str:
    '''Extract student code from login response

    Parameters
    ----------
    login_response : Union[Response, dict]
        The response of the login request.

    Returns
    -------
    str 
        Student code. Represented by :class:`pymyku.attribute.Student.STD_CODE`
    '''

    login_response = response_to_json(login_response)

    student_data = extract_student_data(login_response)

    result = get_raise(student_data, 'stdCode', 'login response')

    return result


def extract_std_id(login_response: Union[Response, dict]) -> str:
    '''Extract student id from login response.

    Parameters
    ----------
    login_response : Union[Response, dict]
        The response of the login request.

    Returns
    -------
    str
        Student id. Represented by :class:`pymyku.attribute.Student.STD_ID`
    '''

    login_response = response_to_json(login_response)

    student_data = extract_student_data(login_response)

    result = get_raise(student_data, 'stdId', 'login response')

    return result


def extract_schedule(schedule_response: Union[Response, dict],
                     as_dict: Optional[bool] = False,
                     full_result: Optional[bool] = False) -> Union[tuple, dict, list]:
    '''Extract academic_year and semester from schedule response.

    Parameters
    ----------
    schedule_response : Union[Response, dict]
        The response from the schedule request.
    as_dict : Optional[bool]
        Whether to return the result as a dictionary or not. By default False
    full_result : Optional[bool]
        If True, return the full result (List[dict]), otherwise return the first item. By default False

    Returns
    -------
    Union[tuple, dict, list]
        Academic year and semester. Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR` and :class:`pymyku.attribute.Schedule.SEMESTER`
    '''

    schedule_response = response_to_json(schedule_response)

    result = get_raise(schedule_response, 'results', 'schedule response')

    if full_result:
        return result

    if isinstance(result, list):
        result = result[0]

    if as_dict:
        return result

    academic_year = get_raise(result, 'academicYr', 'schedule response')

    semester = get_raise(result, 'semester', 'schedule response')

    return academic_year, semester


def gen_request_headers(access_token: Optional[Union[str, Response, dict]] = '') -> dict:
    '''Generate request headers.
    
    Sets the access token as `x-access-token` if provided.

    Parameters
    ----------
    access_token : Optional[Union[str, Response, dict]]
        If the parameter is type Response or dict, the access token will be extracted.
        Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    Returns
    -------
    dict
        Request headers for MyKU API.
    '''

    header = {
        'app-key': constant.APP_KEY,
    }

    if isinstance(access_token, (Response, dict)):
        access_token = extract_access_token(access_token)

    if isinstance(access_token, str) and access_token:
        header['x-access-token'] = access_token

    return header


def gen_login_request_params(username: str, password: str) -> dict:
    '''Generate request parameters for posting login request.

    Parameters
    ----------
    username : str
        Nontri account username (b##########)
    password : str
        Nontri account password (Don't worry, your password is not saved)

    Returns
    -------
    dict
        Parameters for login request.
    '''

    return {
        'url': url.LOGIN,
        'data': {
            'username': username,
            'password': password,
        },
        'headers': {
            'app-key': constant.APP_KEY
        }
    }


def __check_required_kwargs(kwargs: dict, required_kwargs: Union[list, tuple]) -> bool:
    '''Check if all required kwargs are present.

    Parameters
    ----------
    kwargs : dict
        Dictionary of keyword arguments to check.
    required_kwargs : Union[list, tuple]
        Required keyword arguments.

    Returns
    -------
    bool
        True if all required kwargs are present, False otherwise.

    Raises
    ------
    ValueError
        If required kwargs are not present.
    '''

    for kwarg in required_kwargs:
        if not kwargs.get(kwarg) and not str(kwargs.get(kwarg)) == '0':
            raise ValueError(f'{kwarg} is required')
    return True


def gen_request_args_f(function: callable,
                       raise_exception: Optional[bool] = True,
                       **kwargs) -> Dict[str, any]:
    '''Generate request parameters requied for the given function.

    Passing only `login_response`, `schedule_response` or `client` is also acceptable for some request parameters.

    Parameters
    ----------
    function : callable
        Any request function from pymyku.request module.
    raise_exception : Optional[bool]
        Whether to raise exception or not, by default True

    Returns
    -------
    Dict[str, any]
        Request parameters.
        
    Raises
    ------
    ValueError
        If required kwargs are not present.
    '''

    if not callable(function):
        raise ValueError('function must be a callable')

    name = function.__name__

    if kwargs.get('client'):
        client = kwargs['client']
        kwargs['login_response'] = client.get(FetchedResponses.LOGIN_RESPONSE)
        kwargs['schedule_response'] = client.get(
            FetchedResponses.SCHEDULE_RESPONSE)

    if kwargs.get('login_response'):
        login_response = kwargs.get('login_response')
        login_response = response_to_json(login_response)

        user_data = extract_user_data(login_response)

        student_data = user_data['student']

        if not kwargs.get('access_token'):
            kwargs['access_token'] = extract_access_token(login_response)

        kwargs['user_type'] = user_data['userType']

        for key, value in constant.STUDENT_PARAM_DICT.items():
            if not kwargs.get(key):
                kwargs[key] = student_data.get(value)

    if kwargs.get('schedule_response'):
        schedule_response = kwargs.get('schedule_response')
        schedule_response = response_to_json(schedule_response)

        schedule = extract_schedule(schedule_response, as_dict=True)

        if not kwargs.get('academic_year'):
            kwargs['academic_year'] = schedule['academicYr']

        if not kwargs.get('semester'):
            kwargs['semester'] = schedule['semester']

    if raise_exception and not kwargs.get('access_token'):
        raise ValueError('access_token is required')

    headers = gen_request_headers(kwargs.get('access_token'))

    if name == 'logout':

        return {
            'url': url.LOGOUT,
            'headers': headers,
        }

    elif name == 'get_schedule':

        if raise_exception:
            __check_required_kwargs(kwargs, [
                'student_status_code', 'campus_code', 'faculty_code', 'major_code',
                'user_type'
            ])

        return {
            'url': url.SCHEDULE,
            'headers': headers,
            'params': {
                'stdStatusCode': kwargs.get('student_status_code'),
                'campusCode': kwargs.get('campus_code'),
                'facultyCode': kwargs.get('faculty_code'),
                'majorCode': kwargs.get('major_code'),
                'userType': kwargs.get('user_type')
            }
        }

    elif name == 'get_group_course':

        if raise_exception:
            __check_required_kwargs(kwargs, ['academic_year', 'semester', 'std_id'])

        return {
            'url': url.GROUP_COURSE,
            'headers': headers,
            'params': {
                "academicYear": kwargs.get('academic_year'),
                "semester": kwargs.get('semester'),
                "stdId": kwargs.get('std_id')
            }
        }
    elif name == 'get_check_grades':

        return {
            'url': url.CHECK_GRADES,
            'headers': headers,
            'params': {
                'stdCode': kwargs.get('std_code')
            }
        }

    elif name == 'get_gpax':

        return {
            'url': url.GPAX,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }

    elif name == 'get_announce':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id', 'academic_year', 'semester'])

        return {
            'url': url.ANNOUNCE,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id'),
                'academicYear': kwargs.get('academic_year'),
                'semester': kwargs.get('semester')
            }
        }
    elif name == 'search_enroll':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id', 'academic_year', 'semester'])

        return {
            'url': url.SEARCH_ENROLL,
            'headers': headers,
            'data': {
                'stdid': kwargs.get('std_id'),
                'academicYear': str(kwargs.get('academic_year')),
                'semester': str(kwargs.get('semester')),
            }
        }
    elif name == 'get_student_personal':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_PERSONAL,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'get_student_education':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_EDUCATION,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'get_student_address':

        if raise_exception:
            __check_required_kwargs(kwargs, ['std_id'])

        return {
            'url': url.STUDENT_ADDRESS,
            'headers': headers,
            'params': {
                'stdId': kwargs.get('std_id')
            }
        }
    elif name == 'search_subject':

        if raise_exception:
            __check_required_kwargs(kwargs, ['query'])

        return {
            'url': url.SEARCH_SUBJECT,
            'headers': headers,
            'params': {
                'query': kwargs.get('query'),
            }
        }
    elif name == 'search_subject_open':

        if raise_exception:
            __check_required_kwargs(kwargs,
                                    ['query', 'campus_code', 'academic_year', 'semester'])

        return {
            'url': url.SEARCH_SUBJECT_OPEN,
            'headers': headers,
            'params': {
                'query': kwargs.get('query'),
                'section': kwargs.get('section'),
                'campusCode': kwargs.get('campus_code'),
                'academicYear': kwargs.get('academic_year'),
                'semester': kwargs.get('semester')
            }
        }
    elif name == 'search_section_detail':

        if raise_exception:
            __check_required_kwargs(kwargs, ['section_id'])

        return {
            'url': url.SEARCH_SECTION_DETAIL,
            'headers': headers,
            'params': {
                'sectionId': kwargs.get('section_id')
            }
        }



from . import attribute, exception, requests, utils
from .type import (Any, ClientType, Dict, Enum, EnumMeta, List, Optional, Response, Union)


class Client(ClientType):
    '''Represents a client connection that connects to MyKU. This class is used to interact with the MyKU API.
    The client is initialized with a username and password and will login to MyKU automatically.
    '''

    def __init__(self, username: str, password: str) -> None:
        self.__username: str = username
        self.__password: str = password
        self.__login_response: Optional[dict] = {}
        self.__access_token: Optional[str] = ''
        self.__schedule_response: Optional[dict] = {}
        self.__academic_year: Optional[str] = None
        self.__semester: Optional[str] = None

        self.initialize()

    @property
    def login_response(self) -> dict:
        '''Recent login response.

        Returns
        -------
        dict
            Response from the login request. Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
        '''
        return self.__login_response

    @property
    def access_token(self) -> str:
        '''Recent access token from the login response.

        Returns
        -------
        str
            Access token. Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
        '''
        return self.__access_token

    @property
    def schedule_response(self) -> dict:
        '''Recent schedule response.

        Returns
        -------
        Response
            Response from the schedule request. Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
        '''
        return self.__schedule_response

    @property
    def academic_year(self) -> Optional[str]:
        '''Current academic year.

        Returns
        -------
        Optional[str]
            Current academic year. Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        '''
        return self.__academic_year

    @property
    def semester(self) -> Optional[str]:
        '''Current semester.

        Returns
        -------
        Optional[str]
            Current semester. Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        '''
        return self.__semester

    @property
    def user_data(self) -> dict:
        '''Return the user data extracted form login response.

        Returns
        -------
        dict
            User data. Represented by :class:`pymyku.attribute.User`
        '''
        return self.login_response.get('user', {})

    @property
    def student_data(self) -> dict:
        '''Return the student data extracted form login response.

        Returns
        -------
        dict
            Student data. Represented by :class:`pymyku.attribute.Student`
        '''
        return self.user_data.get('student', {})

    @property
    def std_code(self) -> str:
        '''Return the student code extracted from the login response.

        Returns
        -------
        str
            Student code. Represented by :class:`pymyku.attribute.Student.STD_CODE`
        '''
        return self.student_data.get('stdCode')

    def valid_response(self,
                       response: Response,
                       to_json: Optional[dict] = True) -> Union[dict, Response]:
        '''If the response is not 200, check if the error is due to an expired token. 
        If so, reset the token and raise an error. Otherwise, raise the error

        Parameters
        ----------
        response : Response
            The response object returned from the request.
        to_json : Optional[dict]
            If True, the response will be converted to a JSON object.

        Returns
        -------
        Union[dict, Response]
            Return the response if the response is 200. Otherwise, raise the error.

        Raises
        ------
        exception.TokenExpired
            If the response is due to an expired token.
        exception.HTTPError
            If the response is not 200.
        '''

        data = response.json()

        if response.status_code != 200:

            if data.get('code') == 'expired':
                self.reset()
                raise exception.TokenExpired(
                    "The access token has expired. Please reinitialize the client with .initialize()."
                )

        response.raise_for_status()

        return data if to_json else response

    def initialize(self) -> None:
        '''Initialize the client by logging in and fetch user data.
        :meth:`login` will be called to fetch login data.
        After that, :meth:`fetch_schedule` will be called to fetch schedule data.
        
        Affected attributes: :attr:`login_response`, :attr:`.access_token`, :attr:`.schedule_response`, :attr:`.academic_year` and :attr:`.semester`
        '''

        self.login()

        self.__schedule_response = self.fetch_schedule()

        self.__academic_year, self.__semester = utils.extract_schedule(
            self.__schedule_response)

    def reset(self) -> None:
        '''Reset the client attributes.
        
        Affected attributes: :attr:`login_response`, :attr:`.access_token`, :attr:`.schedule_response`, :attr:`.academic_year` and :attr:`.semester`
        '''
        self.__login_response = {}
        self.__access_token = ''
        self.__schedule_response = {}
        self.__academic_year = None
        self.__semester = None

    def headers(self) -> dict:
        '''Return the headers for the requests.

        Returns
        -------
        dict
            The headers for the requests containing :attr:`pymyku.constant.APP_KEY` and :attr:`access_token`.
        '''
        return utils.gen_request_headers(self.__access_token)

    def login(self) -> Response:
        '''Login to MyKu and fetch user data.

        API: https://myapi.ku.th/auth/login
        
        Returns
        -------
        Response
            Response object from the login request. Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
        '''

        login_response = requests.login(self.__username, self.__password)

        login_response.raise_for_status()

        self.__login_response = login_response.json()
        self.__access_token = self.__login_response['accesstoken']

        return login_response

    def fetch_schedule(self,
                       as_response: Optional[bool] = False) -> Union[dict, Response]:
        '''Send GET request to MyKU common/getschedule API.

        API: https://myapi.ku.th/common/getschedule
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request. Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
        '''

        response = requests.get_schedule(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_announce(
            self,
            academic_year=None,
            semester=None,
            as_response: Optional[bool] = False) -> Union[List[dict], Response]:
        '''Send GET request to MyKU advisor/getAnnounceStd API.
                
        API: https://myapi.ku.th/advisor/getAnnounceStd
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            Union[List[dict], Response]

        '''

        if academic_year is None:
            academic_year = self.__academic_year

        if semester is None:
            semester = self.__semester

        response = requests.get_announce(academic_year=academic_year,
                                         semester=semester,
                                         login_response=self.__login_response,
                                         schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def fetch_grades(self, as_response: Optional[bool] = False) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/checkGrades API.

        API: https://myapi.ku.th/std-profile/checkGrades
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request.
        '''

        response = requests.get_check_grades(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_group_course(self,
                           as_response: Optional[bool] = False) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getGroupCourse API.

        API: https://myapi.ku.th/std-profile/getGroupCourse
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request.
        '''

        response = requests.get_group_course(login_response=self.__login_response,
                                             schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_address(self,
                              as_response: Optional[bool] = False
                             ) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdAddress API.

        API: https://myapi.ku.th/std-profile/getStdAddress

        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request.
        '''

        response = requests.get_student_address(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_personal(self,
                               as_response: Optional[bool] = False
                              ) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdPersonal API.

        API: https://myapi.ku.th/std-profile/getStdPersonal
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
            Union[dict, Response]
    
        '''

        response = requests.get_student_personal(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_student_education(self,
                                as_response: Optional[bool] = False
                               ) -> Union[dict, Response]:
        '''Send GET request to MyKU std-profile/getStdEducation API.

        API: https://myapi.ku.th/std-profile/getStdEducation
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request.
        '''

        response = requests.get_student_education(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_gpax(self, as_response: Optional[bool] = False) -> Union[dict, Response]:
        '''Send GET request to MyKU stddashboard/gpax API.
        
        API: https://myapi.ku.th/stddashboard/gpax
        
        Parameters
        ----------
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False

        Returns
        -------
        Union[dict, Response]
            Response from the request.

        '''

        response = requests.get_gpax(login_response=self.__login_response)

        if as_response:
            return response

        return response.json()

    def fetch_enroll(self,
                     academic_year: Optional[Union[str, int]] = None,
                     semester: Optional[Union[str, int]] = None,
                     as_response: Optional[bool] = False) -> Union[dict, Response]:
        '''Send GET request to MyKU enroll/searchEnrollResult API.

        API: https://myapi.ku.th/enroll/searchEnrollResult
        
        Parameters
        ----------
        academic_year : Optional[Union[str, int]], optional
            Academic year, if not provided, will use the current academic year.
            Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        semester : Optional[Union[str, int]], optional
            Semester, if not provided, will use the current semester.
            Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        as_response : Optional[bool]
            Return as Response object if True, otherwise dict, by default False
            
        Returns
        -------
        Union[dict, Response]
            Response from the request.
        '''

        if isinstance(semester, int):
            semester = str(semester)

        if isinstance(academic_year, int):
            academic_year = str(academic_year)

        response = requests.search_enroll(academic_year=academic_year,
                                          semester=semester,
                                          login_response=self.__login_response,
                                          schedule_response=self.__schedule_response)

        if as_response:
            return response

        return response.json()

    def search_subject(self, query: str) -> List[Dict[str, str]]:
        '''Query subject with subject id by sending GET request to MyKU enroll/searchSubjectOpenEnr API.
        
        API: https://myapi.ku.th/enroll/searchSubjectOpenEnr

        Parameters
        ----------
        query : str
            Subject id or name to query. (At least 3 characters), e.g. '013', '01355119', 'eng' or 'english'

        Returns
        -------
        List[Dict[str, str]]
            List of subject dicts.

        Raises
        ------
        ValueError
            The query is less than 3 characters long.
        
        exception.HTTPError
            The request is not successful.
        '''

        if len(query) < 3:
            raise ValueError(
                "Subject ID must be at least 3 characters long.")

        response = requests.search_subject(query=query,
                                           login_response=self.__login_response)

        response = self.valid_response(response)

        return response.get('subjects', [])

    def search_subject_open(
            self,
            subject_id: str,
            section: Optional[str] = '') -> List[Dict[str, Union[str, int]]]:
        '''Query subject enrollment info (All section) of current semester by sending GET request to MyKU enroll/openSubjectForEnroll API.

        API: https://myapi.ku.th/enroll/openSubjectForEnroll
        
        Parameters
        ----------
        subject_id : str
            Subject id to query, e.g. '01355119' or '01355119-64'
        section : Optional[str]
            Section of the subject, e.g. '1'

        Returns
        -------
        List[Dict[str, Union[str, int]]]
            List of subject's opening section.

        Raises
        ------
        Value
            The subject_id is less than 3 characters long.
        exception.HTTPError
            The request is not successful.
        '''

        response = requests.search_subject_open(
            query=subject_id,
            section=section,
            login_response=self.__login_response,
            schedule_response=self.__schedule_response)

        response = self.valid_response(response)

        return response.get('results', [])

    def search_section_detail(
            self, subject_id: str) -> Dict[str, Union[Dict[str, List[Union[str, int]]], list, str]]:
        '''Send GET request to MyKU enroll/searchSectionDetail API.

        API: https://myapi.ku.th/enroll/searchSectionDetail

        Parameters
        ----------
        section_id : str
            Section id. e.g. '186426'

        Returns
        -------
        Dict[str, Union[str, int]]
            _description_
        '''
        
        response = requests.search_section_detail(section_id=subject_id,
                                                  access_token=self.__access_token)
        
        response = self.valid_response(response)
        
        return response.get('sectionDetail', {})

    def get(self, attr: Enum) -> Any:
        '''Get any value from MyKU client. (login response and schedule response)
        Use enums from pymyku.attribute as key to get value.

        Parameters
        ----------
        attr : Enum
            Enum from attribute.

        Returns
        -------
        Any
            Value of the attribute.
            
        Raises
        -------
        TypeError
            The attr is not an Enum.    
        '''

        if isinstance(attr, attribute.FetchedResponses):
            if attr.value == 0:
                return self.__login_response
            elif attr.value == 1:
                return self.__schedule_response

        if isinstance(attr, attribute.Schedule):
            return utils.extract(self.__schedule_response, attr)

        return utils.extract(self.__login_response, attr)

    def get_group_course(self) -> List[dict]:
        '''Send GET request to MyKU std-profile/getGroupCourse API and return the result.

        Returns
        -------
        List[dict]
            List of group course dicts (timetable). 
            
        Raises
        -------
        exception.TokenExpired
            The token is expired.
        exception.HTTPError
            The request is not successful.
        '''

        response = self.fetch_group_course(as_response=True)

        response = self.valid_response(response)

        return response.get('results', [])

    def get_gpax(self) -> float:
        '''Send GET request to MyKU std-profile/getGPAX API and return the gpax.

        Returns
        -------
        float
            GPAX.
            
        Raises
        -------
        exception.TokenExpired
            The token is expired.
        exception.HTTPError
            The request is not successful.
        '''

        response = self.fetch_gpax(as_response=True)

        response = self.valid_response(response)

        return response.get('results', {})[0].get('gpax')

    def get_total_credit(self) -> int:
        '''Send GET request to MyKU std-profile/getTotalCredit API and return the total credit.

        Returns
        -------
        int
            Total credit.
            
        Raises
        ------
        exception.TokenExpired
            The token is expired.
        exception.HTTPError
            The request is not successful.
        '''

        response = self.fetch_gpax(as_response=True)

        response = self.valid_response(response)

        return response.get('results', {})[0].get('total_credit', None)

    def get_grades(self,
                   key: Optional[str] = "subject_code") -> Dict[str, Dict[str, str]]:
        '''Fetch grades for each subjects in all semesters.
        
        Parameters
        ----------
        key : str
            Subject key from the response. 
            Can be `subject_code` or `subject_name_en`.
            Otherwise, `subject_code` will be used.
            Defaults to "subject_code".

        Returns
        -------
        Dict[str, Dict[str, str]]:
            Dict of grades for each subjects in all semesters.
            
        Raises
        -------
        exception.TokenExpired
            The token is expired.
        exception.HTTPError
            The request is not successful.
        '''
        grades = {}

        key = key if key in ("subject_code", "subject_name_en") else "subject_code"

        response = self.fetch_grades(as_response=True)

        response = self.valid_response(response)

        for semester in response['results']:

            year = semester['academicYear']
            grades[year] = {}
            grades[year]['gpa'] = semester['gpa']
            grades[year]['cr'] = semester['cr']
            for subject in semester['grade']:
                grades[year][subject[key]] = subject['grade']

        return grades

    def get_enrolled_subjects(self,
                              academic_year: Optional[Union[str, int]] = None,
                              semester: Optional[Union[str, int]] = None) -> List[dict]:
        '''Get enrolled subjects in a specific semester.

        Parameters
        ----------
        academic_year : Optional[Union[str, int]], optional
            Academic year, if not provided, will use the current academic year.
            Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        semester : Optional[Union[str, int]], optional
            Semester, if not provided, will use the current semester.
            Represented by :class:`pymyku.attribute.Schedule.SEMESTER`

        Returns
        -------
        List[dict]
            List of enrolled subjects.
        '''
        response = self.fetch_enroll(academic_year, semester)
        
        response = self.valid_response(response)

        subjects = [subj for subj in response["enrollSubjects"]]

        return subjects


#: API for logging in to MyKU. API response -- :ref:`:attr:`url.login``
LOGIN: str = 'https://myapi.ku.th/auth/login'
#: API for logging out from MyKU (unsure if this works). API response -- :ref:`:attr:`url.logout``
LOGOUT: str = 'https://myapi.ku.th/auth/logout'


#: API for getting schedule data that contains academic year and semester. API response -- :ref:`:attr:`url.schedule``
SCHEDULE: str = 'https://myapi.ku.th/common/getschedule'

#: API for getting announcement data. API response -- :ref:`:attr:`url.announce``
ANNOUNCE: str = 'https://myapi.ku.th/advisor/getAnnounceStd'

#: API for getting grades for each semester. API response -- :ref:`:attr:`url.check_grades``
CHECK_GRADES: str = 'https://myapi.ku.th/std-profile/checkGrades'
#: API for getting group course. API response -- :ref:`:attr:`url.group_course``
GROUP_COURSE: str = 'https://myapi.ku.th/std-profile/getGroupCourse'
#: API for getting student address information. API response -- :ref:`:attr:`url.student_address``
STUDENT_ADDRESS: str = 'https://myapi.ku.th/std-profile/getStdAddress'
#: API for getting student personal information. API response -- :ref:`:attr:`url.student_personal``
STUDENT_PERSONAL: str = 'https://myapi.ku.th/std-profile/getStdPersonal'
#: API for getting student academic information. API response -- :ref:`:attr:`url.student_education``
STUDENT_EDUCATION: str = 'https://myapi.ku.th/std-profile/getStdEducation'

#: API for getting gpx. API response -- :ref:`:attr:`url.gpax``
GPAX: str = 'https://myapi.ku.th/stddashboard/gpax'

#: API for getting student enrollment data. API response -- :ref:`:attr:`url.search_enroll``
SEARCH_ENROLL: str = 'https://myapi.ku.th/enroll/searchEnrollResult'
#: API for searching subject by subject code. API response -- :ref:`:attr:`url.search_subject``
SEARCH_SUBJECT: str = 'https://myapi.ku.th/enroll/searchSubjectOpenEnr'
#: API for searching opening subject by subject code. API response -- :ref:`:attr:`url.search_subject_open``
SEARCH_SUBJECT_OPEN: str = 'https://myapi.ku.th/enroll/openSubjectForEnroll'
#: API for getting section detail by section id. API response -- :ref:`:attr:`url.search_section_detail``
SEARCH_SECTION_DETAIL: str = 'https://myapi.ku.th/enroll/searchSectionDetail'



from requests import get
from requests import post

from . import utils
from .type import ClientType, Optional, Response, Union


def login(username: str, password: str) -> Response:
    '''Send POST request to MyKU auth/login API.

    API: https://myapi.ku.th/auth/login
    
    Parameters
    ----------
    username : str
        Your Nontri account username (b##########)
    password : str
        Your password (Don't worry, your password is not saved)
    
    Returns
    -------
    Response
        Response object from auth/login API.
    '''

    return post(**utils.gen_login_request_params(username, password))


def logout(access_token: Optional[str] = '',
                   login_response: Optional[Union[Response, dict]] = {},
                   client: Optional[ClientType] = None) -> Response:
    '''Send POST request to MyKU auth/logout API.
    
    *I am uncertain that this request method will work properly.*

    Assigning only `login_response` or `client` is acceptable.
    
    API: https://myapi.ku.th/auth/logout
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from auth/logout API.

    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(logout, **locals())

    return post(**params)


def get_schedule(access_token: Optional[str] = '',
                 user_type: Optional[str] = '',
                 campus_code: Optional[str] = '',
                 faculty_code: Optional[str] = '',
                 major_code: Optional[str] = '',
                 student_status_code: Optional[str] = '',
                 login_response: Optional[Union[Response, dict]] = {},
                 client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU common/getschedule API.

    Assigning only `login_response` or `client` is acceptable.
    
    API: https://myapi.ku.th/common/getschedule

    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    user_type : Optional[str]
        'userType' attribute from login response, Represented by :class:`pymyku.attribute.User.USER_TYPE`
    campus_code : Optional[str]
        'campusCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.CAMPUS_CODE`, e.g. 'B'
        , e.g. 'B'
    faculty_code : Optional[str]
        'facultyCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.FACULTY_CODE`
    major_code : Optional[str]
        'majorCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.MAJOR_CODE`
    student_status_code : Optional[str]
        'studentStatusCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.STUDENT_STATUS_CODE`
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from common/getschedule API.

    Raises
    ------
    ValueError
        Required parameters are missing.
    '''
    params = utils.gen_request_args_f(get_schedule, **locals())

    return get(**params)


def get_group_course(access_token: Optional[str] = '',
                     std_id: Optional[str] = '',
                     academic_year: Optional[str] = '',
                     semester: Optional[str] = '',
                     login_response: Optional[Union[Response, dict]] = {},
                     schedule_response: Optional[Union[Response, dict]] = {},
                     client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU std-profile/getGroupCourse API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.

    API: https://myapi.ku.th/std-profile/getGroupCourse
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    academic_year : Optional[str]
        'academicYr' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        , e.g. '2565'
    semester : Optional[str]
        'semester' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        , e.g. '0' for summer, '1' for first semester and '2' for second semester.
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    schedule_response : Optional[Union[Response, dict]]
        Schedule response from :meth:`get_schedule`, 
        Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from std-profile/getGroupCourse API.

    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_group_course, **locals())

    return get(**params)


def get_check_grades(access_token: Optional[str] = '',
                     std_code: Optional[str] = '',
                     login_response: Optional[Union[Response, dict]] = {},
                     client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU std-profile/checkGrades API.

    Assigning only `login_response` or `client` is acceptable.
    
    
    API: https://myapi.ku.th/std-profile/checkGrades
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_code : Optional[str]
        'stdCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_CODE`
        , e.g. '64xxxxxxxx'
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from std-profile/checkGrades API.
        
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_check_grades, **locals())

    return get(**params)


def get_gpax(access_token: Optional[str] = '',
             std_id: Optional[str] = '',
             login_response: Optional[Union[Response, dict]] = {},
             client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU stddashboard/gpax API.
    
    Assigning only `login_response` or `client` is acceptable.

    API: https://myapi.ku.th/stddashboard/gpax
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from stddashboard/gpax API.

    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_gpax, **locals())

    return get(**params)


def get_announce(access_token: Optional[str] = '',
                 std_id: Optional[str] = '',
                 academic_year: Optional[str] = '',
                 semester: Optional[str] = '',
                 login_response: Optional[Union[Response, dict]] = {},
                 schedule_response: Optional[Union[Response, dict]] = {},
                 client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU advisor/getAnnounceStd API.
    
    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.
    
    API: https://myapi.ku.th/advisor/getAnnounceStd

    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    academic_year : Optional[str]
        'academicYr' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        , e.g. '2565'
    semester : Optional[str]
        'semester' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        , e.g. '0' for summer, '1' for first semester and '2' for second semester.
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    schedule_response : Optional[Union[Response, dict]]
        Schedule response from :meth:`get_schedule`, 
        Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from advisor/getAnnounceStd API.

    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_announce, **locals())

    return get(**params)


def search_enroll(access_token: Optional[str] = '',
                  std_id: Optional[str] = '',
                  academic_year: Optional[str] = '',
                  semester: Optional[str] = '',
                  login_response: Optional[Union[Response, dict]] = {},
                  schedule_response: Optional[Union[Response, dict]] = {},
                  client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU enroll/searchEnrollResult API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.

    API: https://myapi.ku.th/enroll/searchEnrollResult
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    academic_year : Optional[str]
        'academicYr' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        , e.g. '2565'
    semester : Optional[str]
        'semester' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        , e.g. '0' for summer, '1' for first semester and '2' for second semester.
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    schedule_response : Optional[Union[Response, dict]]
        Schedule response from :meth:`get_schedule`, 
        Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from enroll/searchEnrollResult API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.   
    '''

    params = utils.gen_request_args_f(search_enroll, **locals())

    return post(**params)


def get_student_personal(access_token: Optional[str] = '',
                         std_id: Optional[str] = '',
                         login_response: Optional[Union[Response, dict]] = {},
                         client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdPersonal API.

    Assigning only `login_response` or `client` is acceptable.

    API: https://myapi.ku.th/std-profile/getStdPersonal  

    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from std-profile/getStdPersonal API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_student_personal, **locals())

    return get(**params)


def get_student_education(access_token: Optional[str] = '',
                          std_id: Optional[str] = '',
                          login_response: Optional[Union[Response, dict]] = {},
                          client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdEducation API.

    Assigning only `login_response` or `client` is acceptable.

    API: https://myapi.ku.th/std-profile/getStdEducation
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from std-profile/getStdEducation API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''
    params = utils.gen_request_args_f(get_student_education, **locals())

    return get(**params)


def get_student_address(access_token: Optional[str] = '',
                        std_id: Optional[str] = '',
                        login_response: Optional[Union[Response, dict]] = {},
                        client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU std-profile/getStdAddress API.

    Assigning only `login_response` or `client` is acceptable.

    API: https://myapi.ku.th/std-profile/getStdAddress  
    
    Parameters
    ----------
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    std_id : Optional[str]
        'stdId' attribute from login response, Represented by :class:`pymyku.attribute.Student.STD_ID` 
        , e.g. '20xxxx'
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from std-profile/getStdAddress API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(get_student_address, **locals())

    return get(**params)


def search_subject(query: str,
                   access_token: Optional[str] = '',
                   login_response: Optional[Union[Response, dict]] = {},
                   client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU enroll/searchSubjectOpenEnr API.

    Assigning only `login_response` or `client` is acceptable.

    API: https://myapi.ku.th/enroll/searchSubjectOpenEnr  
    
    Parameters
    ----------
    query : str
        Subject id, e.g. '013' or '01355119'
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from enroll/searchSubjectOpenEnr API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(search_subject, **locals())

    return get(**params)


def search_subject_open(query: str,
                        section: Optional[str] = '',
                        access_token: Optional[str] = '',
                        campus_code: Optional[str] = '',
                        academic_year: Optional[str] = '',
                        semester: Optional[str] = '',
                        login_response: Optional[Union[Response, dict]] = {},
                        schedule_response: Optional[Union[Response, dict]] = {},
                        client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU enroll/openSubjectForEnroll API.

    Assigning only (`login_response` and `schedule_response`) or `client` is acceptable.

    API: https://myapi.ku.th/enroll/openSubjectForEnroll  
    
    Parameters
    ----------
    query : str
        Full subject id, e.g. '01355119' or '01355119-64'
    section : Optional[str]
        Section of the subject, e.g. '1'
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    campus_code : Optional[str]
        'campusCode' attribute from login response, Represented by :class:`pymyku.attribute.Student.CAMPUS_CODE`, e.g. 'B'
        , e.g. 'B'
    academic_year : Optional[str]
        'academicYr' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.ACADEMIC_YEAR`
        , e.g. '2565'
    semester : Optional[str]
        'semester' attribute from schedule response, Represented by :class:`pymyku.attribute.Schedule.SEMESTER`
        , e.g. '0' for summer, '1' for first semester and '2' for second semester.
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    schedule_response : Optional[Union[Response, dict]]
        Schedule response from :meth:`get_schedule`, 
        Represented by :class:`pymyku.attribute.FetchedResponses.SCHEDULE_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from enroll/openSubjectForEnroll API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(search_subject_open, **locals())

    return get(**params)


def search_section_detail(section_id: str,
                          access_token: Optional[str] = '',
                          login_response: Optional[Union[Response, dict]] = {},
                          client: Optional[ClientType] = None) -> Response:
    '''Send GET request to MyKU enroll/searchSectionDetail API.

    API: https://myapi.ku.th/enroll/searchSectionDetail
    
    Parameters
    ----------
    section_id : str
        Section id. e.g. '186426'
    access_token : Optional[str]
        'accesstoken' from login response, Represented by :class:`pymyku.attribute.Token.ACCESS_TOKEN`
    login_response : Optional[Union[Response, dict]]
        Login response from login request, can be obtained from :meth:`login`
        Represented by :class:`pymyku.attribute.FetchedResponses.LOGIN_RESPONSE`
    client : Optional[ClientType]
        Initialized :class:`pymyku.Client` object

    Returns
    -------
    Response
        Response object from enroll/searchSectionDetail API.
    
    Raises
    ------
    ValueError
        Required parameters are missing.
    '''

    params = utils.gen_request_args_f(search_section_detail, **locals())

    return get(**params)
